#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstdint>
#include <cstdlib>
#include <unistd.h>
#include <signal.h>
#include <sys/mman.h>
#include <libdis.h>

using namespace std;

/* some global options */

/* verbosity level. Affects the detail of the output */
int verbose = 0;
/* perform a sleep after the execution of each instruction,
 * to emulate the running time on the original machine
 */
useconds_t dosleep = 0;
/* use '@' and '.' instead of '1' and '0' when dumping the
 * contents of memory
 */
bool dotdash = false;
/* show the contents of memory after each instruction, in emulation mode,
 * of after each tb, in binary translation mode
 */
bool animate = false;
/* during animation, reset the screen before each dump */
bool scrnrst = false;
/* enable binary translation */
bool bintrns = false;


/* forward declararations */
struct tb_t;
struct cache_t;

/* the mbaby structure contains the virtual machine state that
 * corresponds to the target machine state. For our Manchester
 * Baby Machine this is just the instruction pointer (CI),
 * the accumulator (A) and the memory (M).
 */
struct mbaby {
	static const int MSIZE = 32;

	int32_t CI;
	int32_t A;
	int32_t M[MSIZE];

	/* constructor: set everything to zero */
	mbaby(): A(0), CI(0) {
		for (int i = 0; i < MSIZE; i++)
			M[i] = 0;
	}

	/* start emulation. The emulation stops at an halt
	 * instruction (opcode 0x7) or, if bintrns is 1, at the first jump
	 * (opcodes 0x0 and 0x01).  Returns 1 if emulation stopped because
	 * of halt, 0 otherwise.
	 */
	int emulate();
	/* load a program from standard input into memory */
	void load();
	/* dump the contents of memory */
	void dump();
	/* start execution using emulation */
	void exec();
	/* start execution using binary translation */
	void btexec();
	/* translate the current dynamic basic block. Returns the
	 * corresponding translated block
	 */
	tb_t* translate();
};

/* for output only: print instruction in symbolic form */
void disas(int32_t instr)
{
	int32_t opcode = (instr & 0xE000) >> 13;
	int32_t addr = instr & 0x1FFF;
	char spaces[] = "                         ";
	ostringstream s;

	switch(opcode) {
	case 0x0:
		s << "CI = [" << addr << "]";
		break;
	case 0x1:
		s << "CI = CI + [" << addr << "]";
		break;
	case 0x2:
		s << "A = -[" << addr << "]";
		break;
	case 0x3:
		s << "[" << addr << "] = A";
		break;
	case 0x4:
	case 0x5:
		s << "A = A - [" << addr << "]";
		break;
	case 0x6:
		s << "if A < 0, CI = CI + 1";
		break;
	case 0x7:
		s << "halt";
		break;
	}
	int nsp = sizeof(spaces) - s.str().size() - 1;
	if (nsp < 0)
		nsp = 0;
	spaces[nsp] = '\0';
	cout << s.str() << spaces;
}

/* restore the terminal cursor (used when animate is active) */
void showcursor(int signum) {
	cout << "\033[?25h";
	exit(0);
}

/* exec using emulation */
void mbaby::exec() {

	/* emulate() assumes that CI has already been incremented from
	 * the previous step, so we need to increment it a first time
	 * at the beginning
	 */
	CI++;
	/* if we are here, bintrns is 0, so emulate() will continue until
	 * if finds an halt instruction
	 */
	emulate();
}

#define MAXHOSTCODE 4096

/* translated block.
 * The translated block is initialized with the address of the first guest
 * instruction. Then, the translation for each guest instruction must
 * be added using addinstr(). Additional code (e.g., prologue and epilogue)
 * may be added with addcode().
 *
 * The code is initially put into a temporary buffer, then moved into the
 * cache when the install() method is called.
 */
struct tb_t {
	/* the size of the temporary buffer */
	static const int TMPCODE = 512;

	/* range of guest address of the dynamic basic block */
	int32_t g_beg;
	int32_t g_end;

	/* range of bytes that contain the translated block.
	 * These initially point into the temporary buffer. When
	 * install() is called, they point into the cache
	 */
	uint8_t *h_beg;
	uint8_t *h_end;

	/* where to put the next bytes of code */
	uint8_t *h_cur;

	/* init a new tb for guest code that starts at ci */
	tb_t(int32_t ci);

	~tb_t();

	/* return true iff the tb contains the guest address g_a */
	bool contains(int32_t g_a) { return g_a >= g_beg && g_a < g_end; }

	/* add code (not related to any guest instruction) to the tb */
	void addcode(uint8_t const* code, int size);

	/* convenience helper to avoid passing the size of the code array */
	template <int N> void addcode(uint8_t const (&code)[N]) { addcode(code, N); }

	/* add the translation for another guest instruction */
	void addinstr(uint8_t const* code, int size);

	/* convenience helper to avoid passing the size of the code array */
	template <int N> void addinstr(uint8_t const (&code)[N]) { addinstr(code, N); }

	/* patch the code added by the last addinstr(). put a at given offset */
	void patch(int offset, int32_t a);

	/* move the generated code to the nextcode pointer and return the pointer
	 * to the first unused byte
	 *
	 * No other code may be added later.
	 */
	uint8_t* install(uint8_t* nextcode);

	/* exec the tb */
	int exec(mbaby *env, uint32_t cache_bitmap)
	{
		/* we cast the h_beg address into a pointer to a function,
		 * then we call the function
		 *
		 * The function receives a pointer to mbaby, which can
		 * be used to reach the emulated registers and memory of the
		 * target machine, and a cache_bitmap, which can be used
		 * to detect self-modifying code
		 */
		typedef int (*tb_f)(mbaby *env, uint32_t);
		return (reinterpret_cast<tb_f>(h_beg))(env, cache_bitmap);
	}

	/* show the contents of the TB on stdout */
	void dump();
};

tb_t::tb_t(int32_t ci):
	h_cur(NULL),
	g_beg(ci),
	g_end(ci)
{
	/* h_beg and h_end initially point into the temporary buffer */
	h_beg = h_end = new uint8_t[TMPCODE];
}

tb_t::~tb_t()
{
	/* we use u_cur==NULL to signal the fact that we have moved the
	 * code from the temporary buffer to the cache and the temporary
	 * buffer has already been deleted
	 */
	if (h_cur)
		delete[] h_beg;
}

void tb_t::addcode(const uint8_t *c, int n)
{
	if ((h_end - h_beg) + n > TMPCODE) {
		cerr << "TMPCODE too small! (need " <<
				(h_end - h_beg) + n <<
				" have " << TMPCODE << ")\n";
		exit(1);
	}
	/* copy the code into the temporary buffer starting at the end.
	 * Move the h_end pointer to the new end
	 */
	while (n--)
		*h_end++ = *c++;
}

void tb_t::addinstr(const uint8_t *c, int n)
{
	/* we are adding the translation of a guest instruction.
	 * In addition to copying the code into the temporary buffer
	 * we also need to let h_cur point to the beginning of the
	 * new code, so that patch() below may use it,
	 * and update the range of our dynamic basic block
	 */
	h_cur = h_end;
	addcode(c, n);
	g_end++;
}

void tb_t::patch(int o, int32_t a)
{
	int32_t *p = reinterpret_cast<int32_t *>(h_cur + o);
	*p = a;
}

uint8_t *tb_t::install(uint8_t *nextcode)
{
	/* we move the code into the cache and delete the
	 * temporary buffer
	 */
	uint8_t *old = h_beg;
	h_beg = nextcode;
	for (uint8_t *scan = old; scan != h_end; scan++)
		*nextcode++ = *scan;
	h_end = nextcode;
	delete(old);
	/* we set h_cur to NULL to remember that the code
	 * is now into the cache
	 */
	h_cur = NULL;
	return nextcode;
}

void tb_t::dump()
{
	cout << "DBB: [" << g_beg << ", " << g_end - 1 << "]\n";
	cout << "TB:\n";
	int col = 0;
	cout << hex << setfill('0');
	for (uint8_t *scan = h_beg; scan != h_end; scan++) {
		if (col == 0)
			cout << "    " << setw(4) << (int)(scan - h_beg) << "  ";
		cout << setw(2) << (int)*scan << " ";
		col++;
		if (col == 16) {
			col = 0;
			cout << "\n";
		}
	}
	if (col != 0)
		cout << "\n";
	cout << "Disassembly:\n";
	int len = h_end - h_beg, pos = 0;
	while (pos < len) {
		x86_insn_t insn;
		char line[256];
		int size = x86_disasm(h_beg, len, 0, pos, &insn);
		x86_format_insn(&insn, line, 256, att_syntax);
		cout << "    " << setw(4) << pos << "  " << line << "\n";
		pos += size;
	}
	cout << dec << setfill(' ');
}

/* The cache maps Dynamic Basic Blocks (DBB) to Translated Blocks (TB) */
struct cache_t {
	/* given the start address of a DBB we need to lookup the
	 * corresponding TB in the cache. Since the guest memory is so small,
	 * we can use a very simple data structure: an array with an entry
	 * for each possibile guest address.  If the entry is NULL, we have no
	 * TB for the given DBB. Otherwise, the entry points to the TB.
	 */
	tb_t* cache[32];
	/* to detect self-modifying code we need to know all the guest addresses
	 * for which we are storing a translation. Again, since there are only
	 * 32 possible addresses, we can use a simple bitmap with a bit for
	 * each guest address.
	 */
	uint32_t bitmap;
	/* code points to the beginning of the code memory area,
	 * nextcode to the end */
	uint8_t *code, *nextcode;

	cache_t();

	~cache_t();

	tb_t* find(int32_t CI)
	{
		return cache[CI];
	}

	/* add a new TB to the cache */
	void add(tb_t *tb);
	/* invalidate all the cache */
	void clear();
	/* invalidate only the TBs that contain the given guest addr */
	void invalidate(int32_t addr);
};

/* this is the global cache */
cache_t cache;

cache_t::cache_t(): bitmap(0)
{
	/* we use mmap() to allocate the cache code area, since we need
	 * the execution permission on the corresponding pages
	 */
	code = static_cast<uint8_t*>(mmap(NULL,			// let the system choose the address
				MAXHOSTCODE,			// requested size of the memory area
				PROT_READ|PROT_WRITE|PROT_EXEC, // which permission bits we want
				MAP_SHARED|MAP_ANONYMOUS,	// this area is not backed by a file
				-1,				// no file descriptor
				0				// offset (ignored)
			));
	nextcode = code;

	/* the cache initially contains nothing */
	for (int i = 0; i < 32; i++)
		cache[i] = NULL;
}

cache_t::~cache_t()
{
	clear();
	munmap(code, MAXHOSTCODE);
}

void cache_t::add(tb_t *tb)
{
	 /* We always add code to the end and, when the cache fills
	  * up, we simply invalidate everthing and start over.
	  */
	if ((nextcode - code) + (tb->h_end - tb->h_beg) > MAXHOSTCODE) {
		clear();
	}
	/* copy the code in the mmaped area and make the tb point there */
	nextcode = tb->install(nextcode);
	/* collect the guest locations that are translated by this tb */
	for (int32_t scan = tb->g_beg; scan != tb->g_end; scan++)
		bitmap |= (1U << scan);
	/* install the modified tb in the cache */
	cache[tb->g_beg] = tb;
}

void cache_t::clear()
{
	for (int i = 0; i < 32; i++) {
		delete cache[i];
		cache[i] = NULL;
	}
	bitmap = 0;
	nextcode = code;
}

void cache_t::invalidate(int32_t addr)
{
	/* there may be more than one TB that contains a translation of addr,
	 * so we need to find and invalidate all of them.
	 *
	 * The bitmap is goint to change, so we recompute it along the way.
	 */
	bitmap = 0;
	for (int32_t scan = 0; scan <= addr; scan++) {
		tb_t *tb = cache[scan];

		if (!tb)
			continue;

		if (tb->contains(addr)) {
			cache[scan] = NULL;
			delete tb;
			continue;
		}

		/* we are keeping this TB, so add it to the bitmap */
		for (int i = tb->g_beg; i < tb->g_end; i++)
			bitmap |= (1U << i);
	}
}

/* emulate() is used in emulation mode to run until end, but also
 * in binary translation mode to step out of a DBB that modifies itself.
 */
int mbaby::emulate() {
	for (;;) {

		int32_t PI = M[CI];
		int32_t opcode = (PI & 0xE000) >> 13;
		int32_t addr = PI & 0x1FFF;

		if (scrnrst) {
			/* cursor in home position */
			cout << "\033[0;0f";
		}
		if (animate) {
			dump();
		}

		switch (opcode) {
		case 0x0:
			CI = M[addr];
			/* if bintrns is active we stop at the end of the DBB
			 * (i.e., at the first jump)
			 */
			if (bintrns)
				return 0;
			break;
		case 0x1:
			CI = CI + M[addr];
			/* same as above */
			if (bintrns)
				return 0;
			break;
		case 0x2:
			A = -M[addr];
			break;
		case 0x3:
			/* this instruction modifies memory and it may
			 * potentially change a guest instruction. If bintrns
			 * is active we need to take care of the cache
			 */
			if (bintrns && ((1U << addr) & cache.bitmap))
				cache.invalidate(addr);
			M[addr] = A;
			break;
		case 0x4:
		case 0x5:
			A = A - M[addr];
			break;
		case 0x6:
			if (A < 0) {
				CI = CI + 1;
			}
			/* again, stop here if bintrns is active */
			if (bintrns)
				return 0;
			break;
		case 0x7:
			return 1;
		}

		/* slow down the program if requested */
		if (dosleep) {
			usleep(dosleep);
		}

		/* the Manchester Baby always increments CI before fetching */
		CI++;
	}
}



tb_t* mbaby::translate() {
	/* we create a TB that looks like a C++ function, with the proper
	 * prologue and epilogue. The function has the following signature:
	 *
	 * int tb(mbaby *env, uint32_t cache_bitmap);
	 *
	 * The return value has the following interpretation:
	 * 0 		normal end of DBB
	 * 1		halt instruction
	 * 2		self-modifying code detected
	 */

	/* in the prologue we initialize some registers so that
	 * they point to the usefult parts of the virtual machine
	 * state (CI, A, M and the cache_bitmap)
	 *
	 * We also allocate a local variable (in -4(%ebp)) to hold
	 * the return value
	 */
	static const uint8_t prologue[] = {
		0x55,			// pushl %ebp
		0x89, 0xe5,		// movl %esp, %ebp
		0x83, 0xec, 0x04,	// subl $4, %esp
		0x57,			// pushl %edi
		0x56,			// pushl %esi
		0x53,			// pushl %ebx
		0x51,			// pushl %ecx
		0x8b, 0x7d, 0x08,	// movl 8(%ebp), %edi	# &m == &CI
		0x8d, 0x5f, 0x04,	// leal 4(%edi), %ebx	# &A
		0x8d, 0x77, 0x08,	// leal 8(%edi), %esi	# &M
		0x8b, 0x4d, 0x0c,	// movl 12(%ebp), %ecx  # cache_bitmap
		0xc7, 0x45, 0xfc,
		0x00, 0x00, 0x00, 0x00	// movl $0, -4(%ebp)	# exit reason
	};

	/* int the epilogue we copy the return value in %eax for the caller
	 * and do the standard cleanup
	 */
	static const uint8_t epilogue[] = {
		0x8b, 0x45, 0xfc,	// movl -4(%ebp), %eax
		0x59,			// popl %ecx
		0x5b,			// popl %ebx
		0x5e,			// popl %esi
		0x5f,			// popl %edi
		0xc9,			// leave
		0xc3			// ret
	};

	/* this is used after the translation of each instruction,
	 * to increment the emulated CI
	 */
	static const uint8_t incCI[] = {
		0xff, 0x07		// incl (%edi)
	};

	/* ci is the address of the guest instruction we are translating.
	 * At the beginning of the translation it is equal to the current
	 * value of CI, the guest instraction that the virtual machine is
	 * about to execute
	 */
	int32_t ci = CI;
	/* create a new TB for a DBB that starts at ci */
	tb_t* tb = new tb_t(ci);

	/* add the prologue above */
	tb->addcode(prologue);

	/* append a translation for all instructions up to the first jump.  The
	 * general idea is to have a "template" of the host binary code
	 * equivalent to each guest opcode. The template contains some
	 * "variables" that depend on the value of the 'addr' field in the
	 * specific guest instruction that we are translating.  These variables
	 * are implemented as "empty fields" (zero bytes) inside the template.
	 * The template is first added to TB as-is, then the empty fields are
	 * filled by patching the TB.
	 */
	for (;;) {

		int32_t PI = M[ci];
		int32_t opcode = (PI & 0xE000) >> 13;
		int32_t addr = PI & 0x1FFF;

		switch (opcode) {
			case 0x0: { 	// CI = M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x86,
					0x00, 0x00, 0x00, 0x00,	// movl X(%esi), %eax
					0x89, 0x07		// movl %eax, (%edi)
				};
				tb->addinstr(templ);
				tb->patch(2, addr * 4);	// set the value of X
				goto out;
			}
			case 0x1: {	// CI = CI + M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x86,
					0x00, 0x00, 0x00, 0x00, // movl X(%esi), %eax
					0x01, 0x07              // addl %eax, (%edi)
				};
				tb->addinstr(templ);
				tb->patch(2, addr * 4);	// set the value of X
				goto out;
			}
			case 0x2: {	// A = -M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x86,
					0x00, 0x00, 0x00, 0x00,	// movl X(%esi), %eax
					0xf7, 0xd8,             // negl %eax
					0x89, 0x03            	// movl %eax, (%ebx)
				};
				tb->addinstr(templ);
				tb->patch(2, addr * 4);	// set the value of X
				break;
			}
			case 0x3: {	// M[addr] = A
				/* this instruction modifies memory, so it may
				 * change an instruction which is already in the cache.
				 * The translated code must first check the cache_bitmap
				 * with the destination address. If there is a match,
				 * the affected TBs must be invalidated before the target
				 * exection may continue.
				 *
				 * For simplicity we return early from the TB and we
				 * perform the invalidation in the main loop.
				 * This is inefficient, since the rest of the DBB
				 * (after the current instruction) will then generate
				 * a new TB.
				 */
				static const uint8_t templ[] = {
					0xf7, 0xc1,
					0x00, 0x00, 0x00, 0x00,	 //    testl $Y, %ecx
					0x74, 0x0b,		 //    jz 1f
					0xb8,
					0x02, 0x00, 0x00, 0x00,	 //    movl $2, %eax
					0x59,			 //    popl %ecx
					0x5b,			 //    popl %ebx
					0x5e,			 //    popl %esi
					0x5f,			 //    popl %edi
					0xc9,			 //    leave
					0xc3,			 //    ret
					0x8b, 0x03,            	 // 1: movl (%ebx), %eax
					0x89, 0x86,
					0x00, 0x00, 0x00, 0x00   //    movl %eax, X(%esi)
				};
				tb->addinstr(templ);
				tb->patch(2, 1U << addr);	// set the value of Y
				tb->patch(23, addr * 4);	// set the value of X
				break;
			}
			case 0x4:
			case 0x5: {	// A = A - M[addr]
				static const uint8_t templ[] = {
					0x8b, 0x86,
					0x00, 0x00, 0x00, 0x00, // movl X(%esi), %eax
					0x29, 0x03              // subl %eax, (%ebx)
				};
				tb->addinstr(templ);
				tb->patch(2, addr * 4);	// set the value of X
				break;
			}
			case 0x6: {	// if (A < 0) CI = CI + 1
				static const uint8_t templ[] = {
					0x83, 0x3b, 0x00,	//     cmpl $0, (%ebx)
					0x7d, 0x02,   		//     jge 1f
					0xff, 0x07   		//     incl (%edi)
								// 1:
				};
				tb->addinstr(templ);
				goto out;
			}
			case 0x7: {	// halt
				static const uint8_t templ[] = {
					0xc7, 0x45, 0xfc,
					0x01, 0x00, 0x00, 0x00	// movl $1, -4(%ebp)
				};
				tb->addinstr(templ);
				goto out;
			}
		}

		/* next instruction to translate */
		ci++;
		/* add the code for the increment of CI */
		tb->addcode(incCI);
		/* Exercise: implement the optimization we discussed in the classroom:
		 * only increment CI at the end of the DBB. Be careful with early exits.
		 */
	}
out:
	/* add the epilogue after the translation of all the instructions in the DBB */
	tb->addcode(epilogue);

	if (verbose > 3)
		tb->dump();

	return tb;
}

void mbaby::btexec()
{
	tb_t *tb;

	for (;;) {
		CI++;

		if (scrnrst) {
			/* cursor in home position */
			cout << "\033[0;0f";
		}

		if (animate)
			dump();

		tb = cache.find(CI);
		if (!tb) {
			tb = translate();
			cache.add(tb);
		}
		switch (tb->exec(this, cache.bitmap)) {
		case 1: 	// halt
			return;
		case 2:	{	// self-modifying code
			int32_t PI = M[CI];
			int32_t addr = PI & 0x1FFF;

			if (tb->contains(addr)) {
				/* the TB is trying to modify itself. We need
				 * to continue with emulation until we step out of it
				 */
				if (emulate())
					return;
			} else {
				/* if the TB is only trying to modify another TB,
				 * invalide the cache for 'addr' and continue in
				 * binary translation mode
				 */
				M[addr] = A;
				cache.invalidate(addr);
			}
			break;
		}
		default:	// normal exit
			break;
		}
	}
}

const char *
bin2ascii(int32_t v)
{
	static char buf[33];
	uint32_t v_ = (uint32_t)v;

	for (int i = 0; i < 32; i++) {
		if (dotdash)
			buf[i] = ((v_ & 0x1) ? '@' : '.');
		else
			buf[i] = '0' + (v_ & 0x1);
		v_ >>= 0x1;
	}
	buf[32] = 0;
	return buf;
}

int32_t
ascii2bin(const char *buf)
{
	uint32_t v = 0;

	for (int i = 0; i < 32; i++) {
		v >>= 0x1;
		if (*buf && *buf++ == '1')
			v |= 0x80000000;
	}
	return (int32_t)v;
}

void mbaby::dump()
{
	for (int i = 0; i < 32; i++) {
		if (verbose)
			cout << setw(4) << i << "  ";
		cout << bin2ascii(M[i]);
		if (verbose > 1) {
			cout << " // " << setw(8) << setbase(16) << M[i] << " " <<
			     setw(16) << setbase(10) << M[i] << " " << (CI == i ? '*' : ' ');
			disas(M[i]);
		}
		cout << "\n";
	}
	if (verbose > 2) {
		cout << '\n';
		cout << "A  " << setw(8) << setbase(16) << A << " " <<
		     setw(16) << setbase(10) << A << "    ";
		disas(A);
	        cout << '\n';
		cout << "CI " << setw(8) << setbase(16) << CI << " " <<
		     setw(16) << setbase(10) << CI << '\n';
	}
}

void mbaby::load()
{
	char buf[80];

	for (int i = 0; i < 32; i++) {
		M[i] = fgets(buf, 80, stdin) ? ascii2bin(buf) : 0;
	}
}

void usage(const char *name) {
	cerr << "Usage: " << name << " [-dvns]\n";
}

int
main(int argc, char *argv[])
{
	char ch;
	bool noexec = false;
	mbaby m;

	while ( (ch = getopt(argc, argv, "advns:rb")) != -1) {
		switch (ch) {
		default:
			usage(argv[0]);
			return 1;
		case 'd':
			dotdash = true;
			break;
		case 'v':
			verbose++;
			break;
		case 'n':
			noexec = true;
			break;
		case 's':
			dosleep = atoi(optarg);
			break;
		case 'a':
			animate = true;
			break;
		case 'r':
			scrnrst = true;
			animate = true;
			break;
		case 'b':
			bintrns = true;
			break;
		}
	}
	m.load();
	if (verbose) {
		cout << "Initial content of memory:\n";
		m.dump();
	}
	if (noexec)
		return 0;
	if (scrnrst) {
		/* clear the screen */
		cout << "\033[2J";
		/* execute showcursor if the program is stopped */
		signal(SIGTERM, showcursor);
		signal(SIGINT, showcursor);
		signal(SIGQUIT, showcursor);
		/* hide cursor */
		cout << "\033[?25l";
	}
	if (bintrns)
		m.btexec();
	else
		m.exec();
	if (verbose)
		cout << "Final content of memory:\n";
	m.dump();
	if (scrnrst)
		showcursor(0);
	return 0;
}
